/**
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

package P2PWNC;

import java.io.*;
import java.security.spec.*;
import java.security.*;
import javax.crypto.*;
import java.nio.ByteBuffer;
import java.math.*;
import java.math.BigInteger;
import java.security.interfaces.*;

/**
 * This class provides cryptographic functionality, that is methods for performing
 * encryption, decryption, digital signing and verification, etc. for the P2PWNC protocol.
 * The standard SUNJCE provider is used.
 *
 * @author Pantelis Frangoudis
 * @author Elias Efstathiou
 * @version 0.0.1
 */
class P2PWNCCrypt extends P2PWNCProto{

    /**
     * Default constructor (does nothing in special, since there are no fields to initialize)
     *
     */
    P2PWNCCrypt () {
    }


    /**
     * Reverses the order of the bytes in the input byte array.
     * @param arr byte[]
     * The byte array whose elements will be reversed.
     * @return byte[]
     * The reversed byte array.
     */
    public static byte[] reverseByteOrder(byte arr[]) {
        int i = 0;
        int j = 0;
        byte retval[] = new byte[arr.length];
        for (i=arr.length; i>0; i--) {
            retval[j] = arr[i-1];
            j++;
        }
        return retval;
    }


       /**
        * Converts from internal key representation to Java KeySpec.
        * This method constructs a KeySpec that corresponds to the input data. For examle, for an
        * RSA public key, it will construct and return an RSAPublicKeySpec object whose
        * public modulus will be the BigInteger representation of the key.data field (byte []) and whose
        * public exponent will be 65537 (as specified by the P2PWNC protocol).
        * @param key P2PWNCKey
        * The P2PWNCKey from which a KeySpec will result.
        * @param pub boolean
        * True if key is public and false if it is a secret key.
        * @return KeySpec
        * The resulting KeySpoec object (can be an instance of the RSAPublicKeySpec, RSAPrivateKeySpec,
        * ECPublicKeySpec or ECPrivateKeySpec classes).
        */
        protected KeySpec internalToKeySpec(P2PWNCKey key, boolean pub) {
		RSAPrivateKeySpec rsapriv;
		RSAPublicKeySpec rsapub;
		ECPrivateKeySpec ecpriv;
		ECPublicKeySpec ecpub;

		byte publicModulus[] = new byte[key.datalen / 2];
		byte privateExponent[] = new byte[key.datalen / 2];

		BigInteger publicExponent = new BigInteger(Integer.toString(65537));

		if (key == null) {
			return null;
		}

		switch (key.algorithm) {
			case P2PWNC_ALGO_RSA:
				if (pub) {
					rsapub = new RSAPublicKeySpec(P2PWNCio.signedBytesToBigInteger2(key.data), publicExponent);
					return rsapub;
				}
				else {
                                    	for (int i=0; i<key.datalen / 2; i++) {
                                                //set private key public modulus
						publicModulus[i] = key.data[i];
					}

                                        int j = 0;
					for (int i = key.datalen /2; i<key.datalen; i++) {
                                                //set private exponent
						privateExponent[j] = key.data[i];
                                                j++;
					}

					rsapriv = new RSAPrivateKeySpec(P2PWNCio.signedBytesToBigInteger2(publicModulus), P2PWNCio.signedBytesToBigInteger2(privateExponent));

					return rsapriv;
					//return crt;
				}

			case P2PWNC_ALGO_ECC: default:
				return null;
				//TODO
		} //endsw
	}

        /**
         * Constructs a Key object from the specified P2PWNCKey input parameter. This method calls
         * internalToKeySpec to derive a KeySpec instance from the input P2PWNCKey object and then
         * constructs a Key object using a KeyFactory.
         * @param key P2PWNCKey
         * The P2PWNCKey object from which a Key object will result.
         * @param pub boolean
         * True if key is public and false if it is secret.
         * @return Key
         * The resulting Key object.
         */
        protected Key internalToKey(P2PWNCKey key, boolean pub) {
		KeyFactory kfact;
		try {
			kfact = KeyFactory.getInstance("RSA");
		}
		catch (java.security.NoSuchAlgorithmException e) {
			System.err.println("internalToKey: No such algorithm");
			return null;
		}
		if (pub) {
			try {
				return kfact.generatePublic(this.internalToKeySpec(key, true));
			}
			catch (java.security.spec.InvalidKeySpecException e) {
				System.err.println("internalToKey: Invalid Public Key KeySpec");
				return null;
			}
		}
		else {
			try {
				return kfact.generatePrivate(this.internalToKeySpec(key, false));
			}
			catch (java.security.spec.InvalidKeySpecException e) {
				System.err.println("internalToKey: Invalid Private Key KeySpec");
				return null;
			}
		}
	}


        /**
         * Encrypts or decrypts a message using a specified P2PWNCKey ("key" parameter).
         * @param message byte[]
         * The message to be encrypted/decrypted.
         * @param key P2PWNCKey
         * The P2PWNCKey key that will be used for the encryption/decryption operation. This key is first
         * converted from P2PWNC internal representation to a JCE-specific object (for example
         * RSAPubliKey) using the internalToKey method.
         * @param cipherMode int
         * Indicates whether the message is to be encrypted or decrypted. Legal values are Cipher.ENCRYPT_MODE and
         * Cipher.DECRYPT_MODE.
         * @param pub boolean
         * True if key is public and false if it is secret.
         * @return byte[]
         * The encrypted/decrypted message.
         */
        protected byte[] cryptMessage(byte message[], P2PWNCKey key, int cipherMode, boolean pub) {
		System.err.println("cryptMessage: Encrypting using public key: " + pub);

		Cipher cipher;

		if ( (cipherMode != Cipher.ENCRYPT_MODE) && (cipherMode != Cipher.DECRYPT_MODE) ) {
			System.err.println("cryptMessage: Invalid Cipher Mode");
			return null;
		}

		if (key.algorithm == P2PWNC_ALGO_RSA) {
			RSAPrivateKey privkey = null;
			RSAPublicKey pubkey = null;

			if (pub) {
				pubkey = (RSAPublicKey)this.internalToKey(key, pub);
			}
			else {
				privkey = (RSAPrivateKey)this.internalToKey(key, pub);
			}

			//key.keyDataToBinaryFile("malakia");
//			System.err.println("******");
//			System.err.println( ((RSAPrivateKey)this.internalToKey(key, pub)).getPrivateExponent().bitLength());
//			System.err.println( ((RSAKey)this.internalToKey(key, pub)).getModulus().bitLength());
//			System.err.println( key.data.length);
			try {
				cipher = Cipher.getInstance("RSA/ECB/PKCS1Padding", "SunJCE");
			}
			catch (java.security.NoSuchAlgorithmException e) {
				System.err.println("encryptMessage: No such algorithm");
				return null;
			}
			catch (javax.crypto.NoSuchPaddingException e) {
				System.err.println("encryptMessage: No such padding");
				return null;
			}
			catch (java.security.NoSuchProviderException e) {
				System.err.println("encryptMessage: No such provider - SunJCE");
				return null;
			}
			try {
				//cipher.init(cipherMode, privkey);
				if (pub) {
					cipher.init(cipherMode, pubkey);
				}
				else {
					cipher.init(cipherMode, privkey);
				}
			}
			catch (java.security.InvalidKeyException e) {
				System.err.println("encryptMessage: Invalid key");
				return null;
			}

			try {
				//returns encrypted / decrypted message
				cipher.update(message);
				return cipher.doFinal();
			}
			catch (javax.crypto.BadPaddingException e) {
				System.err.println("encryptMessage: Bad Padding");
				e.printStackTrace();
				return null;
			}
			catch (javax.crypto.IllegalBlockSizeException e) {
				System.err.println("encryptMessage: Illegal Block Size");
				e.printStackTrace();
				return null;
			}
		}
		else {
			//TODO: ECC
			return null;
		}
	}


	//signs a message using key
	//the message is first hashed using SHA-1
	//In fact the hash is signed
        /**
         * Signs a message using the specified key. The operation works as follows; the message is first
         * SHA-1 hashed and the the 20byte hash value is signed. The result is stored in a P2PWNCSignature
         * object.
         * @param key P2PWNCKey
         * The key that will be used for signing.
         * @param message byte[]
         * The original byte array that is to be signed.
         * @param pub boolean
         * True if key is public.
         * @return P2PWNCSignature
         * The resulting P2PWNCSignature structure (contains the signed data).
         */
        protected P2PWNCSignature signMessage(P2PWNCKey key, byte message[], boolean pub) {
		MessageDigest md;
		byte signatureData[];
		byte digest[];

		if (message == null) {
			System.err.println("Null message");
			return null;
		}

		try {
			md = MessageDigest.getInstance("SHA1");
			md.update(message);
			digest = md.digest();
		}
		catch (java.security.NoSuchAlgorithmException e) {
			System.err.println("signMessage: No such algorithm - SHA1");
			return null;
		}

		if (digest == null) {
			System.err.println("Null digest");
			return null;
		}

		signatureData = this.cryptMessage(digest, key, Cipher.ENCRYPT_MODE, pub);
/*		Signature sig = null;
		try {
			sig = Signature.getInstance("SHA1withRSA");
		}
		catch (NoSuchAlgorithmException e) {
			System.err.println("signMessage: No such algorithm - SHA1withRSA");
			return null;
		}


		try {
			sig.initSign( (PrivateKey)this.internalToKey(key, false));
		}
		catch (InvalidKeyException e) {
			System.err.println("signMessage: Invalid Key");
			return null;
		}

		try {
			sig.update(message);
			signatureData = sig.sign();
		}
		catch (SignatureException e) {
			System.err.println("signMessage: Error signing message");
			e.printStackTrace();
			return null;
		}

*/

		if (signatureData == null) {
			System.err.println("Null signature data");
			return null;
		}

		return new P2PWNCSignature(key.algorithm, key.bits, signatureData);
	}

        /**
         * Verifies a P2PWNCSignature. That is, extracts that P2PWNCSignature data, decrypts them, SHA-1 hashes
         * the data the signature has resulted from and compares the hash values.
         * @param signature P2PWNCSignature
         * The P2PWNCSignature that is to be verified.
         * @param key P2PWNCKey
         * The key that will be used for signature verification.
         * @param message byte[]
         * The original data, from which the signature has resulted.
         * @param pub boolean
         * True if key is public.
         * @return boolean
         * True if verification succeeded, false otherwise.
         */
        protected boolean verifySignature(P2PWNCSignature signature, P2PWNCKey key, byte message[], boolean pub) {
		MessageDigest md;
		byte signatureData[];
		byte digest[];

		//first calculate message SHA-1 hash
		try {
			md = MessageDigest.getInstance("SHA1");
			md.update(message);
			digest = md.digest();
		}
		catch (java.security.NoSuchAlgorithmException e) {
			System.err.println("signMessage: No such algorithm - SHA1");
			return false;
		}

		//decrypt signature using key
		signatureData = this.cryptMessage(signature.data, key, Cipher.DECRYPT_MODE, pub);

		//compare decrypted message digest to computed SHA1 hash value
		//if they match ==> signature verified
		return digest.equals(signatureData);

/*
		Signature sig = null;
		try {
			sig = Signature.getInstance("SHA1withRSA");
		}
		catch (NoSuchAlgorithmException e) {
			System.err.println("verifyMessage: No such algorithm - SHA1withRSA");
			return false;
		}


		try {
			sig.initVerify( (PublicKey)this.internalToKey(key, true));
		}
		catch (InvalidKeyException e) {
			System.err.println("verifyMessage: Invalid Key");
			return false;
		}

		try {
			sig.update(message);
			return sig.verify(signature.getData());
		}
		catch (SignatureException e) {
			System.err.println("signMessage: Error signing message");
			e.printStackTrace();
			return false;
		}
*/
	}

        /**
         * Serializes the information contained in a P2PWNC receipt. The result of this function is
         * the byte array that will be signed. It contains the client certificate data (issuer team
         * public key, user public key, team signature - these values are byte arrays and must be serialized
         * strictly in the above order), the service provider public key, the session timestamp as an
         * integer indicating the number of seconds that have elapsed since the Epoch and the receipt weight
         * (an integer value of the traffic uploaded/downloaded from a client during a session, measured in bytes).
         * The above integer values are serialized in big endian form.
         * @param cert P2PWNCCert
         * The client certificate that is included in the receipt.
         * @param prov P2PWNCKey
         * The service provider public key.
         * @param timestamp int
         * The session timestamp (number of seconds since the Epoch).
         * @param weight int
         * Session weight (traffic uploaded/downloaded by a client during the session, measured in bytes).
         * @return byte[]
         */
        protected byte[] serializeReceiptInfo(P2PWNCCert cert, P2PWNCKey prov, int timestamp, int weight) {
		byte retval[];
		int len = 0;
		ByteBuffer buff = ByteBuffer.allocate(8);

		if (cert == null){
			System.err.println("serializeReceiptInfo: Null certificate");
			return null;
		}

		if (prov == null) {
			System.err.println("serializeReceiptInfo: Null provider key");
			return null;
		}

		len =
			cert.teamPublicKey.datalen +
			cert.userPublicKey.datalen +
			cert.teamSignature.datalen +
			prov.datalen +
			8;

		buff.putInt(timestamp);
		buff.putInt(weight);


		retval = new byte[len];
		int pos = 0;

		for (int i=0; i<cert.teamPublicKey.datalen; i++) {
			retval[pos] = cert.teamPublicKey.data[i];
			pos++;
		}
		for (int i=0; i<cert.userPublicKey.datalen; i++) {
			retval[pos] = cert.userPublicKey.data[i];
			pos++;
		}
		for (int i=0; i<cert.teamSignature.datalen; i++) {
			retval[pos] = cert.teamSignature.data[i];
			pos++;
		}
		for (int i=0; i<prov.datalen; i++) {
			retval[pos] = prov.data[i];
			pos++;
		}
		for (int i=0; i<8; i++) {
			retval[pos] = buff.array()[i];
			pos++;
		}

		return retval;
	}

        /**
         * Constructs and signs a P2PWNC receipt. It creates a receipt using the input parameters. Then, it
         * serializes these information (using serializeReceiptInfo), signs the resulting byte array and sets the
         * receipt's receiptSignature field to the resulting object.
         * @param cert P2PWNCCert
         * Client certificate that will be included in the receipt.
         * @param prov P2PWNCKey
         * Service provider public key that will be included in the receipt.
         * @param timestamp int
         * The session timestamp (number of seconds since the Epoch).
         * @param weight int
         * Session weight (traffic uploaded/downloaded by a client during the session, measured in bytes).
         * @param key P2PWNCKey
         * The key that will be used to sign the receipt.
         * @param pub boolean
         * True if key is public.
         * @return P2PWNCReceipt
         * The resulting P2PWNCReceipt object. This can be passed to the receiptToTextMsg so as to
         * create an RCPT text message.
         */
        public P2PWNCReceipt constructReceipt(P2PWNCCert cert, P2PWNCKey prov, int timestamp, int weight, P2PWNCKey key, boolean pub) {
		P2PWNCReceipt rcpt;
		byte rinfo[];

		rinfo = this.serializeReceiptInfo(cert, prov, timestamp, weight);

		//sha-1 hashes and signs receipt info
		P2PWNCSignature rsig = this.signMessage(key, rinfo, pub);
		return new P2PWNCReceipt(cert, prov, timestamp, weight, rsig);
	}

        public static void cryptTest(P2PWNCKey priv, P2PWNCKey pub, byte data[]) throws
            NoSuchAlgorithmException, InvalidKeyException, SignatureException {
            P2PWNCCrypt crypter = new P2PWNCCrypt();
            RSAPrivateKey PRIV = (RSAPrivateKey)crypter.internalToKey(priv, false);
            RSAPublicKey PUB = (RSAPublicKey)crypter.internalToKey(pub, true);
            System.err.println("My keys:\nPrivate:\n" + PRIV.getPrivateExponent().toByteArray()[0]  +
                    " - " + PRIV.getModulus().toByteArray()[0]);
            System.err.println("Public:\n" + PUB.getPublicExponent().toByteArray()[0] + " - " + PUB.getModulus().toByteArray()[0]);
/*
            try {
                    KeyPairGenerator generator = KeyPairGenerator.getInstance("RSA");
                    generator.initialize(1024);
                    KeyPair pair = generator.generateKeyPair();
                    PRIV = (RSAPrivateKey)pair.getPrivate();
                    PUB = (RSAPublicKey)pair.getPublic();
            }
            catch (NoSuchAlgorithmException e) {
            }
            System.err.println("JAVA keys:\nPrivate:\n" + PRIV.getPrivateExponent().toByteArray()[0]  +
                    " - " + PRIV.getModulus().toByteArray()[0]);
            System.err.println("Public:\n" + PUB.getPublicExponent().toByteArray()[0] + "-" + PUB.getModulus().toByteArray()[0]);

            for (int i=0; i<PRIV.getPrivateExponent().toByteArray().length; i++) {
                System.err.print("(" + PRIV.getPrivateExponent().toByteArray()[i] + ")");
            }
            System.err.println("\n" + PRIV.getPrivateExponent().toByteArray().length);
*/

            Signature sig = Signature.getInstance("SHA1withRSA");
            sig.initSign(PRIV);
            sig.update(data);
            byte signed[] = sig.sign();
            Signature sig2 = Signature.getInstance("SHA1withRSA");
            sig2.initVerify(PUB);
            sig2.update(data);
            System.err.println(sig2.verify(signed));

        }


}
