/**
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

package P2PWNC;

import java.io.*;
import java.net.*;
import java.math.BigInteger;

/**
 * This class provides i/o functionality. That is, it provides methods for reading/writing messages from/to
 * a socket, reading keys and certificates from files, etc.
 *
 * @author Pantelis Frangoudis
 * @author Elias Efstathiou
 * @version 0.0.1
 */
class P2PWNCio extends P2PWNCProto {
	P2PWNCio() {}

	//reads certificate from file
	//cert is stored in ascii form
	//first: algorithm+bits
	//then, b64-encoded keys and signature
	public P2PWNCCert readCert(String path) {
		BufferedReader reader;
		String line = null;
		StringBuffer text = new StringBuffer();
		P2PWNCKey teampub = null;
		P2PWNCKey userpub = null;
		P2PWNCSignature teamsig = null ;
		int rawlen = 0;
		Base64 encoder = new Base64();

		int algorithm = 0;
		int bits = 0;
		int crlfpos = 0;

		try {
		 reader = new BufferedReader(new FileReader(path));
		}
		catch (FileNotFoundException e) {
			System.err.println("readCert: File not found");
			return null;
		}

		try {
			//read first line: algorithm + bits
			if ( (line = reader.readLine()) != null) {
				if (line.substring(0, 3).compareToIgnoreCase("RSA") != 0) {
					if (line.substring(0, 3).compareToIgnoreCase("ECC") != 0) {
						System.err.println("Invalid algorithm name or name not found");
						reader.close();
						return null;
					}
					else {
						algorithm = P2PWNC_ALGO_ECC;
					}
				}
				else {
					algorithm = P2PWNC_ALGO_RSA;
				}
			}
			try {
				bits = Integer.parseInt(line.substring(3));
			}
			catch (NumberFormatException e) {
				System.err.println("Error parsing bits field");
				return null;
			}

			//read team pub key
			while ( (line = reader.readLine()) != null) {
				if (line.length() != 0) {
					text.append(line);
				}
				else {
					//crlfcrlf
					rawlen = encoder.CalculateRawSize(text.toString());
					teampub = new P2PWNCKey(algorithm, bits, rawlen);
					encoder.DecodeData(text.toString(), teampub.data);
					text.delete(0, text.length()); //empty string buffer
					break;
				}
			}

			//read user pub key
			while ( (line = reader.readLine()) != null) {
				if (line.length() != 0) {
					text.append(line);
				}
				else {
					//crlfcrlf
					rawlen = encoder.CalculateRawSize(text.toString());
					userpub = new P2PWNCKey(algorithm, bits, rawlen);
					encoder.DecodeData(text.toString(), userpub.data);
					text.delete(0, text.length()); //empty string buffer
					break;
				}
			}

			//read team signature
			while ( (line = reader.readLine()) != null) {
				if (line.length() != 0) {
					text.append(line);
				}
				else {
					break;
				}
			}
			rawlen = encoder.CalculateRawSize(text.toString());
			teamsig = new P2PWNCSignature(algorithm, bits, rawlen);
			encoder.DecodeData(text.toString(), teamsig.data);

		}
		catch (IOException e) {
			System.err.println("Error reading file");
			return null;
		}

		try {
			reader.close();
		}
		catch (IOException e) {
			System.err.println("Error closing input file");
			return null;
		}

		return new P2PWNCCert(teampub, userpub, teamsig);
	}


	//reads a public/private key from an (ascii)file
	//returns a P2PWNCKey object (or null on error)
	public P2PWNCKey readKey(String path) {
		BufferedReader reader;
		String line = null;
		StringBuffer text = new StringBuffer();
		P2PWNCKey key = null;
		int rawlen = 0;
		Base64 encoder = new Base64();

		int algorithm = 0;
		int bits = 0;
		int crlfpos = 0;

		try {
		 reader = new BufferedReader(new FileReader(path));
		}
		catch (FileNotFoundException e) {
			System.err.println("readKey: File not found");
			return null;
		}

		try {
			//read first line: algorithm + bits
			if ( (line = reader.readLine()) != null) {
				if (line.substring(0, 3).compareToIgnoreCase("RSA") != 0) {
					if (line.substring(0, 3).compareToIgnoreCase("ECC") != 0) {
						System.err.println("readKey: Invalid algorithm name or no algorithm identifier");
						reader.close();
						return null;
					}
					else {
						algorithm = P2PWNC_ALGO_ECC;
					}
				}
				else {
					algorithm = P2PWNC_ALGO_RSA;
				}
			}
			try {
				bits = Integer.parseInt(line.substring(3));
			}
			catch (NumberFormatException e) {
				System.err.println("Error parsing bits field");
				return null;
			}

			//read key
			while ( (line = reader.readLine()) != null) {
				if (line.length() != 0) {
					text.append(line);
				}
				else {
					break;
				}
			}
//			System.err.print(text);
			rawlen = encoder.CalculateRawSize(text.toString());
			key = new P2PWNCKey(algorithm, bits, rawlen);
			encoder.DecodeData(text.toString(), key.data);
//			encoder.DecodeFile("aueb_", "decoded");
		}
		catch (IOException e) {
			System.err.println("Error reading file");
			return null;
		}

		try {
			reader.close();
		}
		catch (IOException e) {
			System.err.println("Error closing input file");
			return null;
		}
		//System.err.println("datalen: " + key.datalen);
		return key;
	}

	public P2PWNCKey readKeyBin(String path, int len) throws IOException {
		FileInputStream fin = new FileInputStream(path);
		P2PWNCKey key = new P2PWNCKey(P2PWNC_ALGO_RSA, 1024, len);
		fin.read(key.data);
		fin.close();
		return key;
	}


	public P2PWNCMessage readMessage(Socket socket) {
		BufferedReader reader = null;
		String line = null;
		P2PWNCMessage msg = new P2PWNCMessage();
		StringBuffer textMessage = new StringBuffer();
		int clen = 0; //content length
		char content[];

		try {
			reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));
		}
		catch (IOException e) {
			System.err.println("readMessage: Error obtaining socket input stream");
			return null;
		}

		try {
			//first read message type
			line = reader.readLine();
			if (line == null) {
                            System.err.println("readMessage: Error reading message type");
                            return null;
			}

			if ( line != null) {
				if (line.substring(0, 15).compareToIgnoreCase("CACK P2PWNC/1.0") == 0) {
					msg.mtype = P2PWNC_MSG_CACK;
				}
				else
				if (line.substring(0, 15).compareToIgnoreCase("RREQ P2PWNC/1.0") == 0) {
					msg.mtype = P2PWNC_MSG_RREQ;
				}
				else {
					//unknown message type
                                        System.err.println("readMessage: Unknown message type - " + line.substring(0, 4));
					return null;
				}
			}
			textMessage.append(line + "\r\n");

			//then read content length
			line = reader.readLine();
			if (line == null) {
                            System.err.println("readMessage: Error reading content length");
                            return null;
			}
			if (line != null) {
				if (line.substring(0, "Content-length: ".length()).compareToIgnoreCase("Content-length: ") != 0) {
                                    System.err.println("readMessage: Error reading content length value");
                                    return null;
				}
				else {
					try {
						clen = Integer.parseInt(line.substring("Content-length: ".length()));
					}
					catch (NumberFormatException e) {
						System.err.println("Error parsing content length");
						return null;
					}
				}
			}
			textMessage.append(line + "\r\n");

			//finally, read (the remaining) content-length bytes
			content = new char[clen];
			reader.read(content, 0, clen);
		}
		catch (IOException e) {
			System.err.println("Error reading message");
			return null;
		}
		msg.setText(textMessage + new String(content));
                return msg;
	}

	public static BigInteger signedBytesToBigInteger(byte signed[]) {
		int curByte = 0;
		StringBuffer bitstring = new StringBuffer();
		StringBuffer curstr = new StringBuffer();
		for (int i = 0; i<signed.length; i++) {
			curByte = (int)(signed[i] & 255);
			//convert to bitstring and pad with zero's if bitlen < 8
			for (int j=0;j<(8 - Integer.toBinaryString(curByte).length()); j++) {
				curstr.append("0");
			}
			curstr.append(Integer.toBinaryString(curByte));
			bitstring.append(curstr);
			curstr.delete(0, 8);
		}
		//System.err.println(bitstring.toString());
		//System.err.println("len = " + bitstring.toString().length());
		return new BigInteger(bitstring.toString());
	}


	protected boolean sendTextMsg(Socket socket, String msg) {
//		PrintWriter out;
/*		try {
			out = new PrintWriter(socket.getOutputStream(), true);
		}
		catch (IOException e) {
			System.err.println("sendTextMsg: Could not obtain socket output stream");
			return false;
		}

		out.print(msg);
		out.close();
		return true;
*/
            try {
                socket.getOutputStream().write(msg.getBytes());
            } catch (IOException ex) {
            }
              return true;
	}


	public static BigInteger signedBytesToBigInteger2(byte signed[]) {
                byte newb[] = new byte[signed.length + 1];
                if (signed[0] >= 0) {
                    return new BigInteger(signed);
                }
		System.arraycopy(signed, 0, newb, 1, signed.length);
//		if (signed[0] < 0) newb[0] = 0;
//		else newb[0] = -1;
                newb[0] = 0;
		BigInteger bi = new BigInteger(newb);
                return bi;
	}
}

